local boxinfo = require('cartridge.lua-api.boxinfo')
local cartridge = require('cartridge')
local checks = require('checks')
local confapplier = require('cartridge.confapplier')
local json = require('json')
local log = require('log')
local luri = require('uri')
local msgpack = require('msgpack')
local topology = require('cartridge.lua-api.topology')

local Router = require('ouroboros-cartridge-sidecar.http_router')

local role_name = 'ouroboros-topology'
local role_api_path = '/ouroboros/api/v1'

---@class (exact) SpaceMeta
---@field name string Space name
---@field space string Space tuple encoded to MsgPack and then to hex
---@field indexes string[] Space indexes tuples encoded to MsgPack and then to hex

---@class (exact) ServerInfo
---@field uuid string Server UUID
---@field uri string Server URI
---@field is_leader boolean True if the server is the current master at the time of the request

---@class (exact) SpaceInfo
---@field space_id number Space ID in current replicaset
---@field space_name string Space name
---@field bucket_id_index number Index of bucket_id field in tuple, starting from 0

---@class (exact) ReplicasetInfo
---@field uiud string Replicaset UUID
---@field alias string Replicaset alias
---@field is_storage boolean True if the replicaset has the role 'ouroboros-replicator'
---@field is_router boolean True if the replicaset has the role 'ouroboros-topology'
---@field servers ServerInfo[] List of replicaset servers
---@field spaces SpaceInfo[] List of replicaset spaces

---@class (exact) ShardGroupInfo
---@field name string Shardgroup name
---@field bucket_count number Number of buckets
---@field replicasets ReplicasetInfo[] List of shard group replicasets
---@field bucket_idx_name string Name of field with bucket_id if spaces have format
---@field bucket_idx_number integer Number of index with bucket_id if spaces haven't format, starting from 1

local function render_json(app)
    return function(request)
        local result = app(request)
        if result.json then
            result.headers = result.headers or {}
            result.headers['Content-Type'] = 'application/json'
            result.body = json.encode(result.json)
            result.json = nil
            result.status = result.status or 200
        end
        return result
    end
end

local function render_errors(app)
    return function(request)
        local ok, result = xpcall(function() return app(request) end, function(err)
            if type(err) == 'string' then
                if err:find("Can't decode json in request") then
                    err = {type = 'BAD_REQUEST', message = err}
                elseif err:find("Unknown ") then
                    err = {type = 'NOT_FOUND', message = err}
                else
                    err = {type = 'ERROR', message = err}
                end
            end
            if err.silent then
                return err
            elseif err.type == 'BAD_REQUEST' or err.type == 'NOT_FOUND' or err.type == 'UNPROCESSABLE_ENTITY' then
                log.info(err)
            else
                log.error(err)
            end
            return err
        end)

        if ok then
            return result
        elseif result.type == 'BAD_REQUEST' then
            return {status = 400, json = {error = result.message or 'Invalid request data'}}
        elseif result.type == 'NOT_FOUND' then
            return {status = 404, json = {error = result.message or 'Not found'}}
        elseif result.type == 'UNPROCESSABLE_ENTITY' then
            return {status = 422, json = {error = result.message or 'Request can not be processed'}}
        else
            return {status = 500, json = {error = result.client_message or 'Internal server error'}}
        end
    end
end

-- name of default vshard group if there is only one group
local DEFAULT_GROUP = 'default'

-- function to get spaces from replicaset master server
local SPACES_FUN = [[
local local_fiber = require('fiber')
local json = require('json')
local is_lschema, lschema = pcall(require, 'vshard.storage.schema')
if not is_lschema then
    lschema = nil
end
local vshard = require('vshard')
local spaces = {}
local shard_index = (lschema and lschema.shard_index)
    or vshard.storage.internal.shard_index
    or 'bucket_id'
local i = 0
-- get all spaces with 'bucket_id' field
for _, tuple in box.space._space:pairs() do
    i = i + 1
    if i % 200 == 0 then
        local_fiber.yield()
    end
    if tuple[1] > 512 then -- we don't need system spaces
        local bucket_index
        if type(shard_index) == 'string' then
            bucket_index = box.space._index.index.name:get{tuple[1], shard_index}
        else
            bucket_index = box.space._index.index.primary:get{tuple[1], shard_index}
        end
        if bucket_index and #bucket_index.parts then
            if #bucket_index.parts[1] > 0 then
                table.insert(spaces, {
                    space_name = tuple[3],
                    space_id = tuple[1],
                    bucket_id_index = bucket_index.parts[1][1],
                })
            elseif bucket_index.parts[1].field then
                table.insert(spaces, {
                    space_name = tuple[3],
                    space_id = tuple[1],
                    bucket_id_index = bucket_index.parts[1].field,
                })
            end
        end
    end
end
return spaces, shard_index
]]

-- function to create special user or update his password
local USER_FUN = [[
local new_pass = ...
if new_pass then
    if box.schema.user.exists('ouroboros') then
        box.schema.user.passwd('ouroboros', new_pass)
    else
        box.schema.user.create('ouroboros', { password = new_pass })
        box.schema.user.grant('ouroboros', 'read,write,create,alter,drop,execute,session', 'universe')
    end
end
return true
]]

-- function to get space and its indexes.
-- If space doesn't exist or non-sharded, returns false
local SPACE_INFO_FUN = [[
local space_id_name = ...
if not space_id_name then
    return false
end
local is_lschema, lschema = pcall(require, 'vshard.storage.schema')
if not is_lschema then
    lschema = nil
end
local vshard = require('vshard')
local shard_index = (lschema and lschema.shard_index)
    or vshard.storage.internal.shard_index
    or 'bucket_id'

local space_meta
local _, space_id = pcall(tonumber, space_id_name)
if space_id then
    space_meta = box.space._space:get{space_id}
else
    space_meta = box.space._space.index.name:get{space_id_name}
end
if not space_meta then
    return false
end

local bucket_index
if type(shard_index) == 'string' then
    bucket_index = box.space._index.index.name:get{space_meta[1], shard_index}
else
    bucket_index = box.space._index:get{space_meta[1], shard_index}
end
if not bucket_index or #bucket_index.parts == 0 then
    return false
end

local res = {
    space = space_meta,
    indexes = {}
}
for _, index in box.space._index.index.primary:pairs({space_meta[1]}) do
    table.insert(res.indexes, index)
end

return true, res
]]

--- Collect spaces from replicasets of selected shard group and enrich replicasets info or raise an error
---@param router table Router of the current shard group
---@param replicasets table Replicasets of the current shard group (map of replicaset_uuid -> ReplicasetInfo)
---@return table
---@return string|number
local function get_spaces(router, replicasets)
    local allroutes = router:routeall()
    local shard_index
    for rs_uuid, rs in pairs(replicasets) do
        local ri = allroutes[rs_uuid]
        if ri and rs.is_storage then
            local conn = ri:connect_replica(ri.master)
            local c_res, c_err = conn:wait_connected(30)
            if c_res == nil then
                error(c_err)
            end
            local rs_spaces, r_shard_index, r_err = conn:eval(SPACES_FUN)
            if r_err then
                error(r_err)
            end
            rs.spaces = rs_spaces
            shard_index = shard_index or r_shard_index
        end
    end

    return replicasets, shard_index
end

--- Handle HTTP request: get cluster topology wuth vshard groups, replicasets and spaces
---@return table "HTTP response with json=ShardGroupInfo[]"
local function get_topology()
    local result = {}
    local t_groups = {}

    -- vshard_groups, replicasets and servers
    local cluster_topology = assert(topology.get_topology())
    for replicaset_uuid, replicaset in pairs(cluster_topology.replicasets) do
        local vshard_group = replicaset.vshard_group
        if vshard_group and not t_groups[vshard_group] then
            table.insert(t_groups, vshard_group)
        end
        vshard_group = vshard_group or DEFAULT_GROUP
        if not result[vshard_group] then
            result[vshard_group] = { replicasets = {}, bucket_count = 0 }
        end
        local group_replicasets = result[vshard_group].replicasets
        local replicaset_info = {
            servers = {},
            spaces = {},
            is_storage = false,
            is_router = false,
            vshard_group = vshard_group,
            alias = replicaset.alias,
            http_address = nil,
        }
        local is_storage = false
        local is_router = false
        local replicator = require('ouroboros-cartridge-sidecar.roles.ouroboros-replicator')
        local topology = require('ouroboros-cartridge-sidecar.roles.ouroboros-topology')
        for _, role in ipairs(replicaset.roles) do
            if role == replicator.role_name then
                is_storage = true
            end
            if role == topology.role_name then
                is_router = true
            end
        end
        replicaset_info.is_storage = is_storage
        replicaset_info.is_router = is_router
        for _, server_info in ipairs(replicaset.servers) do
            table.insert(replicaset_info.servers, {
                uuid = server_info.uuid,
                uri = server_info.uri,
                is_active_master = replicaset.active_master.uuid == server_info.uuid,
            })
            if replicaset_info.is_router and replicaset.active_master.uuid == server_info.uuid then
                local ok, info = pcall(boxinfo.get_info, server_info.uri)

                if ok and info then
                    local http_host = info.general.http_host
                    if http_host == '0.0.0.0' then
                        local uri = luri.parse(server_info.uri)
                        http_host = uri.host or uri.ipv4 or uri.ipv6
                    end
                    local http_port = info.general.http_port
                    if http_host and http_port then
                        replicaset_info.http_address = ('%s:%s'):format(http_host, http_port)
                    end
                end
            end
        end
        group_replicasets[replicaset_uuid] = replicaset_info
    end

    -- buckets
    local vshard_groups = confapplier.get_readonly('vshard_groups')
    if not vshard_groups then
        error('Vshard groups config does not exist')
    end
    for group_name, group_info in pairs(vshard_groups) do
        result[group_name].bucket_count = group_info.bucket_count
    end

    -- spaces
    local routers = cartridge.service_get('vshard-router')
    if #t_groups == 0 then
        local replicasets, bucket_id_info = get_spaces(routers.get(), result[DEFAULT_GROUP].replicasets)
        result[DEFAULT_GROUP].replicasets = replicasets
        if type(bucket_id_info) == 'number' then
            result[DEFAULT_GROUP].bucket_idx_name = ''
            result[DEFAULT_GROUP].bucket_idx_number = bucket_id_info
        else
            result[DEFAULT_GROUP].bucket_idx_name = bucket_id_info
            result[DEFAULT_GROUP].bucket_idx_number = 0
        end
    else
        for _, group_name in ipairs(t_groups) do
            local replicasets, bucket_id_info = get_spaces(routers.get(group_name), result[group_name].replicasets)
            result[group_name].replicasets = replicasets
            if type(bucket_id_info) == 'number' then
                result[group_name].bucket_idx_name = ''
                result[group_name].bucket_idx_number = result[group_name].bucket_idx_number or bucket_id_info
            else
                result[group_name].bucket_idx_name = result[group_name].bucket_idx_name or bucket_id_info
                result[group_name].bucket_idx_number = 0
            end
        end
    end

    local res = {}
    for group_name, group_info in pairs(result) do
        local shard_group = {
            name = group_name,
            bucket_count = group_info.bucket_count,
            bucket_idx_name = group_info.bucket_idx_name,
            bucket_idx_number = group_info.bucket_idx_number,
            replicasets = {},
        }
        for r_uuid, ri in pairs(group_info.replicasets) do
            ri.uuid = r_uuid
            table.insert(shard_group.replicasets, ri)
        end
        table.insert(res, shard_group)
    end

    return {json = res}
end

--- Create a special user on all replicasets with role 'ouroboros-cartridge-sidecar.roles.ouroboros-replicator'
--- or update password if user does exist
---@param router table Router of the current shard group
---@param replicasets table Replicasets of the current shard group (map of replicaset_uuid -> ReplicasetInfo)
---@param new_pass string User password
local function create_update_user(router, replicasets, new_pass)
    local allroutes = router:routeall()
    for r_uuid, replicaset in pairs(allroutes) do
        if replicasets[r_uuid] then
            local conn = replicaset:connect_replica(replicaset.master)
            local c_res, c_err = conn:wait_connected(30)
            if c_res == nil then
                error(c_err)
            end
            -- local ok, conn:eval(USER_FUN, {new_pass})
            local ok, err = pcall(function() return conn:eval(USER_FUN, {new_pass}) end)
            if not ok then
                error(('Error on create/update user, instance %s: %s'):format(r_uuid, err))
            end
        end
    end
end

--- Hadnle HTTP request: create specials user on replicasets if it does not exist
---@param req table HTTP request
---@return table "HTTP response with json = {result = 'ok}"
local function check_user(req)
    local data = req:json()
    if not data.new_pass or type(data.new_pass) ~= 'string' then
        error({type= 'BAD_REQUEST', msg = 'wrong password'})
    end

    local replicasets = {}
    local cluster_topology = assert(topology.get_topology(), 'Can not get cluster topology')
    local replicator = require('ouroboros-cartridge-sidecar.roles.ouroboros-replicator')
    for replicaset_uuid, replicaset in pairs(cluster_topology.replicasets) do
        local is_storage = false
        for _, role in ipairs(replicaset.roles) do
            if role == replicator.role_name then
                is_storage = true
                break
            end
        end
        if is_storage then
            replicasets[replicaset_uuid] = true
        end

    end

    local routers = cartridge.service_get('vshard-router')
    local vshard_groups = assert(confapplier.get_readonly('vshard_groups'), 'Can not get vshard groups')
    for group_name, _ in pairs(vshard_groups) do
        local router
        if group_name == DEFAULT_GROUP then
            router = routers.get()
        else
            router = routers.get(group_name)
        end
        create_update_user(router, replicasets, data.new_pass)
    end
    return {json = {result = 'ok'}}
end

--- Check if 2 space info tables are equal.
--- Space and its indexes may have different IDs on different replicasets so we can not compare
--- tuples content, only number of indexes
---@param etalon SpaceMeta
---@param candidate SpaceMeta
---@return boolean
local function is_tables_equal(etalon, candidate, depth)
    checks('table', 'table', 'number|nil')

    local function is_equal(left, right, depth)
        if type(left) ~= type(right) then
            return false
        end
        if type('left') == 'table' then
            return is_tables_equal(left, right, depth + 1)
        else
            return left == right
        end
    end

    depth = depth or 0
    if depth > 30 then
        return false
    end

    if #etalon ~= #candidate then
        return false
    end
    if #etalon ~= 0 then
        for i = 1, #etalon do
            if not is_equal(etalon[i], candidate[i], depth) then
                return false
            end
        end
    else
        for k, v in ipairs(etalon) do
            if not is_equal(v, candidate[k], depth) then
                return false
            end
        end
        for k, _ in ipairs(candidate) do
            if not etalon[k] then
                return false
            end
        end
    end
    return true
end

--- Handle HTTP request: get metdata and indexes for space id or name.
--- Every replicaset should return non-empty space meta
---@param req table HTTP request
---@return table "HTTP response with json = {result = 'ok}"
local function get_space_meta(req)
    local group = req:stash('group')
    local space = req:stash('space')

    if not (group and space) then
        error("Invalid parameters")
    end

    local vshard_groups = confapplier.get_readonly('vshard_groups')
    if group ~= DEFAULT_GROUP and not vshard_groups[group] then
        error("Unknown vshard group name " .. group)
    end

    local routers = cartridge.service_get('vshard-router')
    local group_router
    if group == DEFAULT_GROUP then
        group_router = routers.get()
    else
        group_router = routers.get(group)
    end

    -- collect space meta from all replicasets
    local allroutes = group_router:routeall()
    local has_no_space_replica = false
    local space_all_metas = {}
    for uuid, replicaset in pairs(allroutes) do
        local conn = replicaset:connect_replica(replicaset.master)
        local c_res, c_err = conn:wait_connected(30)
        if c_res == nil then
            error(c_err)
        end
        local r_ok, res, r_err = conn:eval(SPACE_INFO_FUN, { space })
        if r_err then
            error(r_err)
        end
        if r_ok then
            space_all_metas[uuid] = res
        else
            has_no_space_replica = true
            break
        end
    end

    if has_no_space_replica then
        return { json = {} }
    end

    -- check if all metas are equal
    local not_equal = false
    local res
    for _, info in pairs(space_all_metas) do
        if not res then
            res = info
        else
            not_equal = not_equal or not is_tables_equal(res, info)
        end
    end

    if not_equal or not res then
        return { json = {} }
    else
        local result = {
            name = res.space[3],
            space = string.hex(msgpack.encode(res.space)),
            indexes = {}
        }
        for _, idx in ipairs(res.indexes) do
            table.insert(result.indexes, string.hex(msgpack.encode(idx)))
        end
        return { json = result }
    end
end

local function init()
    local httpd = assert(cartridge.service_get('httpd'), "Failed to get httpd service")

    local router = Router:new(httpd, {prefix = role_api_path, middleware = function(app)
        return render_json(render_errors(app))
    end})

    local instance_self = topology.get_self()
    local health_status = {
        ['service_name'] = 'ouroboros',
        ['service_status'] = 1,
        alias = instance_self.alias,
    }
    router:get('/health', function(req)
        return req:render({ json = health_status })
    end)
    router:get('/version', function(req)
        return req:render({ json = { version = require("ouroboros-cartridge-sidecar.version") } })
    end)

    router:get('/topology', get_topology)
    router:get('/space/:group/:space', get_space_meta)
    router:post('/user', check_user)
end

local function stop()
    local httpd = cartridge.service_get('httpd')
    for i = #httpd.routes, 1, -1 do
        if string.sub(httpd.routes[i].path, 1, string.len(role_api_path)) == role_api_path then
            table.remove(httpd.routes, i)
            break
        end
    end
end

local function validate_config()
    return true
end

local function apply_config()
    return true
end

local function get_issues()
    local issues = {}
    return issues
end

return {
    role_name = role_name,
    init = init,
    stop = stop,
    validate_config = validate_config,
    apply_config = apply_config,
    get_issues = get_issues,
    dependencies = {
        'cartridge.roles.vshard-router',
    },
}