
require("os").setenv("TARANTOOL_LOG", "")
local t = require('luatest')
local fio = require('fio')
local cartridge_helpers = require('cartridge.test-helpers')
local requests = require('test.requests')
local topology = require('ouroboros-cartridge-sidecar.roles.ouroboros-topology')
local replicator = require('ouroboros-cartridge-sidecar.roles.ouroboros-replicator')

local app_root = fio.dirname(fio.abspath(package.search('init')))
local helper = {
    shared = {
        root = app_root,
        datadir = fio.pathjoin(app_root, 'tmp', 'db_test'),
    }
}

t.before_suite(function()
    fio.rmtree(helper.shared.datadir)
    fio.mktree(helper.shared.datadir)
end)

helper.cluster = cartridge_helpers.Cluster:new({
    server_command = fio.pathjoin(helper.shared.root, 'init.lua'),
    datadir = helper.shared.datadir,
    use_vshard = true,
    replicasets = {
        {
            alias = 'topology',
            uuid = t.helpers.uuid('a'),
            roles = {
                'test.roles.router',
                topology.role_name,
            },
            servers = {
                {instance_uuid = t.helpers.uuid('a', 1), alias = 'topology1'},
            },
        },
        {
            alias = 'replicaset-1',
            uuid = t.helpers.uuid('b'),
            roles = {
                'test.roles.storage-1',
                replicator.role_name,
            },
            servers = {
                {instance_uuid = t.helpers.uuid('b', 1), alias = 'storage-1-1'},
                {instance_uuid = t.helpers.uuid('b', 2), alias = 'storage-1-2'},
            },
            vshard_group = 'test',
        },
        {
            alias = 'replicaset-2',
            uuid = t.helpers.uuid('c'),
            roles = {
                'test.roles.storage-2',
                replicator.role_name,
            },
            servers = {
                {instance_uuid = t.helpers.uuid('c', 1), alias = 'storage-2-1'},
                {instance_uuid = t.helpers.uuid('c', 2), alias = 'storage-2-2'},
                {instance_uuid = t.helpers.uuid('c', 3), alias = 'storage-2-3'},
            },
            vshard_group = 'new',
        },
    },
})

function helper.start_cluster()
    print('start cluster')
    helper.cluster:start()
    -- Increase default vshard call timeout when luacov is enabled (because it slows down requests)
    local _, runner = pcall(require, 'luacov.runner')
    if runner and runner.initialized then
        for _, server in pairs(helper.cluster.servers) do
            server.net_box:eval('require("vshard").consts.CALL_TIMEOUT_MIN = 1')
        end
    end

    requests.setup({
        port = helper.cluster.main_server.http_port,
    })
end

function helper.get_servers_by_role_name(cluster, role_name)
    local replicas = {}

    for _, replicaset in ipairs(cluster.replicasets) do
        for _, role in ipairs(replicaset.roles) do
            if role == role_name then
                table.insert(replicas, replicaset.uuid)
                break
            end
        end
    end

    local result = {}
    for _, server in pairs(cluster.servers) do
        for _, replicaset in ipairs(replicas) do
            if server.replicaset_uuid == replicaset then
                table.insert(result, server)
            end
        end
    end

    return result
end

function helper.is_server_readonly(server)
    return assert(server.net_box):eval('return box.cfg.read_only')
end

function helper.get_master_by_role_name(cluster, role_name)
    for _, server in ipairs(helper.get_servers_by_role_name(cluster, role_name)) do
        if not helper.is_server_readonly(server) then
            return server
        end
    end
    return nil
end

t.before_suite(function()
    fio.rmtree(helper.shared.datadir)
    fio.mktree(helper.shared.datadir)
    helper.start_cluster()
end)

t.after_suite(function()
    helper.cluster:stop()
    print('cluster is stopped')
end)

return helper
