local t = require('luatest')
local g = t.group('ouroboros_replicator')
local helper = require('test.helper.integration')
local msgpack = require('msgpack')
local requests = require('test.requests')

local table_data

g.test_ext_lib = function()
    local replicator = require('ouroboros-cartridge-sidecar.roles.ouroboros-replicator')
    local instance = assert(helper.get_master_by_role_name(helper.cluster, replicator.role_name))
    local res = instance.net_box:call('libouroboros_test_lib')
    t.assert(res)
end

local to_bytes = function(data)
    local packed = msgpack.encode(data)
    local res = {}
    for i = 1, #packed do
        table.insert(res, string.byte(string.sub(packed, i, i)))
    end
    return res
end

g.test_create_space = function()
    local replicator = require('ouroboros-cartridge-sidecar.roles.ouroboros-replicator')
    local instance = assert(helper.get_master_by_role_name(helper.cluster, replicator.role_name))
    -- check if space doesn't exist
    local res = instance.net_box:eval("return box.space.test_space_create == nil")
    t.assert_equals(res, true)
    -- create space
    local space_name = 'test_space_create'
    local data1 = {
        1024,
        32,
        space_name,
        'memtx',
        0,
        {is_sync = false},
        {
            {
                name = 'id',
                type = 'unsigned',
            },
        },
    }
    local data1_bytes = to_bytes(data1)
    local key = {}
    local key_bytes = to_bytes(key)
    local create_msg = {
        280,
        key_bytes,
        data1_bytes,
        'Replace',
        msgpack.NULL,
    }
    local res, err = instance.net_box:call('libouroboros_apply_ddl', {create_msg})
    t.assert_not(err)
    t.assert_equals(res, true)
    -- check space
    res = instance.net_box:eval("return box.space.test_space_create ~= nil")
    t.assert_equals(res, true)
    -- replace format
    local data2 = {
        1024,
        32,
        space_name,
        'memtx',
        0,
        {is_sync = false},
        {
            {
                name = 'id',
                type = 'unsigned',
            },
            {
                name = 'name',
                type = 'string',
            },
        },
    }
    local data2_bytes = to_bytes(data2)
    local key = {data2[1]}
    local key_bytes = to_bytes(key)
    local update_msg = {
        280,
        key_bytes,
        data2_bytes,
        'Replace',
        space_name,
    }
    local res, err = instance.net_box:call('libouroboros_apply_ddl', {update_msg})
    t.assert_not(err)
    t.assert_equals(res, true)
    -- check space
    res = instance.net_box:eval("return box.space.test_space_create ~= nil")
    t.assert_equals(res, true)
    -- check format
    res = instance.net_box:eval("return box.space._space:get({box.space.test_space_create.id})")
    t.assert(res)
    t.assert(res[7])
    t.assert_equals(#res[7], 2)
end

g.test_drop_space = function()
    local replicator = require('ouroboros-cartridge-sidecar.roles.ouroboros-replicator')
    local instance = assert(helper.get_master_by_role_name(helper.cluster, replicator.role_name))
    table_data = instance.net_box:eval([[
local test_space = box.schema.space.create('test_space_delete')
return box.space._space:get(test_space.id)
    ]])
    local key = {table_data[1]}
    local key_bytes = to_bytes(key)
    local drop_msg = {
        280,
        key_bytes,
        {},
        'Delete',
        'test_space',
    }
    local res, err = instance.net_box:call('libouroboros_apply_ddl', {drop_msg})
    t.assert_not(err)
    t.assert_equals(res, true)
    -- check space
    res = instance.net_box:eval("return box.space.test_space ~= nil")
    t.assert_equals(res, false)
    -- repeated DELETE DDL should not fail
    res, err = instance.net_box:call('libouroboros_apply_ddl', {drop_msg})
    t.assert_not(err)
    t.assert_equals(res, true)
end

g.test_dml = function()
    local replicator = require('ouroboros-cartridge-sidecar.roles.ouroboros-replicator')
    local instance = assert(helper.get_master_by_role_name(helper.cluster, replicator.role_name))
    -- create test space
    local res = instance.net_box:eval([[
local test_space = box.schema.space.create('test_space_dml', {format = {{name = 'id', type = 'unsigned'}, {name = 'name', type = 'string'}}})
test_space:create_index('primary', {parts = {'id'}})
return true
]])
    t.assert_equals(res, true)
    -- create record
    local space_name = 'test_space_dml'
    local data1 = {
        1,
        'name 1',
    }
    local data1_bytes = to_bytes(data1)
    local create_msg = {
        space_name,
        data1_bytes
    }
    local res, err = instance.net_box:call('libouroboros_apply_dml_replace', {create_msg})
    t.assert_not(err)
    t.assert_equals(res, true)
    -- check records
    res = instance.net_box:eval("return box.space.test_space_dml:get{1}")
    t.assert_equals(res, data1)
    -- check replace existed record
    res, err = instance.net_box:call('libouroboros_apply_dml_replace', {create_msg})
    t.assert_not(err)
    t.assert_equals(res, true)
    -- replace name
    local data2 = {
        1,
        'another name',
    }
    local data2_bytes = to_bytes(data2)
    local update_msg = {
        space_name,
        data2_bytes
    }
    res, err = instance.net_box:call('libouroboros_apply_dml_replace', {update_msg})
    t.assert_not(err)
    t.assert_equals(res, true)
    -- check records
    res = instance.net_box:eval("return box.space.test_space_dml:get{1}")
    t.assert_equals(res, data2)
    -- delete record
    local key = { data1[1] }
    local key_bytes = to_bytes(key)
    local delete_msg = {
        space_name,
        key_bytes
    }
    local res, err = instance.net_box:call('libouroboros_apply_dml_delete', {delete_msg})
    t.assert_not(err)
    t.assert_equals(res, true)
    res = instance.net_box:eval("return box.space.test_space_dml:get{1}")
    t.assert_not(res)
end

local function get_storage_master_slave()
    local cluster = helper.cluster;

    local replicators = helper.get_servers_by_role_name(cluster, require('ouroboros-cartridge-sidecar.roles.ouroboros-replicator').role_name)

    local replicaset_uuid
    local master
    local slave

    for _, server in ipairs(replicators) do
        if not replicaset_uuid then
            replicaset_uuid = server.replicaset_uuid
        end
        if replicaset_uuid == server.replicaset_uuid then
            local readonly = helper.is_server_readonly(server)
            if readonly and slave == nil then
                slave = server
            end

            if not readonly and master == nil then
                master = server
            end
        end
        if master ~= nil and slave ~= nil then break end
    end

    return master, slave
end

g.test_metrics = function()
    local master, slave = get_storage_master_slave()
    t.assert_not_equals(master.http_port, slave.http_port)

    -- master metrics should have alias and uuid
    local master_requests = requests.new({ port = master.http_port, prefix = '/' })
    local master_metrics = master_requests:metrics().body
    t.assert_str_contains(master_metrics, "replicaset_alias=")
    t.assert_str_contains(master_metrics, "replicaset_uuid=")

    -- slave metrics should not have alias and uuid
    local slave_requests = requests.new({ port = slave.http_port, prefix = '/' })
    local slave_metrics = slave_requests:metrics().body
    t.assert_not_str_contains(slave_metrics, "replicaset_alias=")
    t.assert_not_str_contains(slave_metrics, "replicaset_uuid=")

    -- making master non-electable, so slave becomes a master
    master:graphql({
        query = [[
            mutation editTopology($replicasets: [EditReplicasetInput!], $servers: [EditServerInput!]) {
                cluster {
                    edit_topology(replicasets: $replicasets, servers: $servers) {
                        servers {
                            uuid
                        }
                    }
                }
            }
        ]],
        variables = {
            servers = {{
                uuid = master.instance_uuid,
                electable = false
            }}
        }
    })

    -- confirm that failover happened
    t.assert_eval_to_true(helper.is_server_readonly(master))
    t.assert_eval_to_false(helper.is_server_readonly(slave))

    -- master metrics now should NOT have alias and uuid
    local master_requests = requests.new({ port = master.http_port, prefix = '/' })
    local master_metrics = master_requests:metrics().body
    t.assert_not_str_contains(master_metrics, "replicaset_alias=")
    t.assert_not_str_contains(master_metrics, "replicaset_uuid=")

    -- slave metrics now should have alias and uuid
    local slave_requests = requests.new({ port = slave.http_port, prefix = '/' })
    local slave_metrics = slave_requests:metrics().body
    t.assert_str_contains(slave_metrics, "replicaset_alias=")
    t.assert_str_contains(slave_metrics, "replicaset_uuid=")
end